
comment		^

The Theory:

~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Routine LZW_COMPRESS

    set w = NIL
loop
  read a character k
  if wk exists in the dictionary
      w = wk
  else
      output the code for w
      add wk to the dictionary
      w = k
endloop

~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Routine LZW_DECOMPRESS
read a character k
output k
w = k
loop
   read a character 
   entry = dictionary entry for k 
   output entry (fully translate all codes to final characters)
   add w + first char of entry to the dictionary (first char in above string value -0255)
   w = entry (previous inputted k)
endloop

~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


       -Our dictionary will be 4096 - 24 bit entries. To utilize REP SCASD
	primitives it will actually be 4096 DWORDS with the upper 8 bits ignored.

       -Max String size in Dictionary is 2 bytes (24 bits)
		^
.data

	cDictionary = 4096	;entries in dictionary
	pDictionary	dd	?
	Temp		dd	?
	FirstChar	dd	?		;first char in entry

.code
	Recurse		proto, :dword
	ReadByIndex 	proto, :dword, :dword
	WriteByIndex	proto, :dword, :dword

;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
	LZWCompress proc, Source: dword, Destination: dword, cbSize: dword
;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

	local	ptrLastByte:	dword

	;--- Create Space for Dictionary and Init entries 0-0xFF ---
	invoke VirtualAlloc, NULL, cDictionary*4, MEM_COMMIT, PAGE_READWRITE
	mov pDictionary, eax

	;--- Init Counters ---
	mov eax, cbSize		;byte size of source file
	add eax, Source		;eax = final byte of source + 1
	mov ptrLastByte, eax

	mov esi, Source		;ptr source file
	xor ebx, ebx		;Destination Counter (index)
	mov ecx, 100h		;index of next available entry in Dictionary
	xor eax, eax		;STRING = NULL
	mov dl, 0		;CHARACTER = NULL

	;--- Load String with first Character (w) ---
	mov al, byte ptr[esi]
	inc esi

.repeat

;Remember, everything must be converted to 12 bit values!
	
	mov dl, byte ptr[esi]	;get input character (k)	
	shl eax, 12		;make room in string for 12 bit character
	and eax,111111111111111111111111b	;clear upper 8 bits (unused)
	mov al, dl		;eax = wk

	;--- Search Dictionary for wk ---
	pushad
	mov edi, pDictionary
	add edi, (4 * 100h) 	;edi = ptr to start of dictionary
	sub ecx, 0100h		;dont search below edi

	.if ecx		;make sure there's at least one entry to search
		cld
		repne scasd
		sub edi, 4
	.else 
		mov edi, -1
	.endif


	.if (edi != -1) && (eax == [edi])	;if found in Dictionary, get code
		
		sub edi, pDictionary	;offset of entry
		shr edi, 2		;/ sizeof DWORD = char code
		mov Temp, edi	
		popad
		mov eax, Temp
		
	.else		;if not found, add to Dictionary and output w
		mov Temp, eax
		popad		
		mov edi, pDictionary
		mov eax, Temp
		.if ecx != cDictionary	;still have room left in Dictionary
			shl ecx, 2	;* sizeof DWORD
			add edi, ecx	;edi = offset new Dictionary entry
			mov [edi], eax	;save wk in Dictionary
			shr ecx, 2	;restore ecx to index value
			inc ecx		;advance Dictionary index
		.endif
			push eax
			shr eax, 12	;isolate w		
			invoke WriteByIndex, Destination, ebx	;send w to output
			inc ebx			;advance Destination ptr index
			pop eax		
	.endif

	
;--- Check for EOF ---
	inc esi
.until esi == ptrLastByte

	and eax, 111111111111b	;isolate k
	invoke WriteByIndex, Destination, ebx	;send final k to output
	push ebx
	invoke VirtualFree, pDictionary, NULL, MEM_RELEASE

	;--- Get return value (byte count of Destination file)
	pop eax		;0 base counter of 12 bit entries saved in compressed file
	inc eax
	mov ebx, eax
	shr ebx, 1	;/2
	pushfd		;save carry flag
	add eax, ebx
	popfd
	.if CARRY?
		inc eax
	.endif

	ret

LZWCompress endp

;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
	LZWDeCompress proc, Source: dword, Destination: dword, cbSize: dword
;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

	local	DictionaryUpdated:	dword

	;--- Create Space for Dictionary and Init entries 0-0xFF ---
	invoke VirtualAlloc, NULL, cDictionary*4, MEM_COMMIT, PAGE_READWRITE
	mov pDictionary, eax

	;--- Init Counters ---
	;iterations = (cbSize * 8 bits) / 12 bits
	mov eax, cbSize
	shl eax, 3	;*8
	mov edx, 0	
	mov ebx, 12	;divisor
	div ebx
	mov cbSize, eax	;number of 12 bit values to decode in source
	

	mov esi, Destination	;ptr Destination file
	xor ebx, ebx		;Source Counter (index)
	mov ecx, 100h		;index of next available entry in Dictionary
	xor eax, eax		;STRING = NULL
	xor edx, edx 		;CHARACTER = NULL

	;--- Load String (w) with first Character (k) ---
	invoke ReadByIndex, Source, ebx
	mov byte ptr[esi], dl
	inc esi
	inc ebx
	shl edx, 12		
	mov eax, edx	;w = k

.repeat
	mov DictionaryUpdated, FALSE		;clear updated flag
	invoke ReadByIndex, Source, ebx		;edx == k
	mov Temp, edx				;w = temp at end of iteration

	;--- If k = next uncreated Dictionary entry, first char of last k (now w)
	;--- appended to w and creates the next Dictionary Entry
	;--- ^^^^ Comparable to the Caramilk Secret ^^^
	.if ecx == edx
		push eax
		push edx
		mov edx, FirstChar	;first char of last string output
		or eax, edx		;eax = new Dictionary entry 
		
		shl ecx, 2	;* sizeof DWORD
		mov edi, pDictionary
		add edi, ecx	;edi = offset new Dictionary entry
		mov [edi], eax	;save wk in Dictionary
		shr ecx, 2	;restore ecx to index value
		inc ecx		;advance Dictionary index
		pop edx
		pop eax
		mov DictionaryUpdated, TRUE	;only update once per iteration
	.endif
	
	mov FirstChar, -1			;invalidate 
	;--- If k is less than 255 then output k ---
	.if edx < 100h		;is char. output k
		mov byte ptr[esi], dl	;send k to output
		mov FirstChar, edx
		inc esi
	
	;--- If k is greater than 255, its a code ---	
	.else			
		;--- Get Dictionary entry for k ---
		mov edi, edx
		shl edi, 2	;* sizeof DWORD
		add edi, pDictionary	;edi = ptr to entry
		mov edi, [edi]		;edi = dictionary entry for k
		invoke Recurse, edi
	.endif
		

	;--- Add w + FirstChar to Dictionary ---
	.if (DictionaryUpdated == FALSE) && (ecx != cDictionary)	
			mov edx, FirstChar
			mov al, dl	;add first char to w
			shl ecx, 2	;* sizeof DWORD
			mov edi, pDictionary
			add edi, ecx	;edi = offset new Dictionary entry
			mov [edi], eax	;save wk in Dictionary
			shr ecx, 2	;restore ecx to index value
			inc ecx		;advance Dictionary index
	.endif
	
	
	;--- Make w = k ---
	mov eax, Temp
	shl eax, 12	;w = k (k= blank)


	;--- Check for EOF ---
	inc ebx
.until ebx == cbSize

	;--- Return byte length of uncompressed file ---
	sub esi, Destination
	mov eax, esi
	ret

LZWDeCompress endp

;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
	Recurse	proc,	Entry: dword
;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

 ;--- Handle First Value of Entry ---
	mov edx, Entry
	shr edx, 12		;edx = first value of Dictionary Entry

	.if edx < 100h		;if first value = char
		mov byte ptr[esi], dl
		inc esi
		.if FirstChar == -1
			mov FirstChar, edx
		.endif
	.else			;first value = code
		
		;--- Get Dictionary entry for code ---
		mov edi, edx
		shl edi, 2	;* sizeof DWORD
		add edi, pDictionary	;edi = ptr to entry
		mov edi, [edi]		;edi = dictionary entry for k
		invoke Recurse, edi
	.endif


 ;--- Handle Second Value of Entry ---

	mov edx, Entry
	and edx, 111111111111b	;isolate second value

	.if edx < 100h		;if char
		mov byte ptr[esi], dl
		inc esi
		ret
	.else			;if code
		;--- Get Dictionary entry for code ---
		mov edi, edx
		shl edi, 2	;* sizeof DWORD
		add edi, pDictionary	;edi = ptr to entry
		mov edi, [edi]		;edi = dictionary entry for k
		invoke Recurse, edi
		ret
	.endif

Recurse endp

;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
	ReadByIndex proc, Source: dword, Index: dword
;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
;Returns 12 bit value in edx (clears upper 20 bits)

;Since we are dealing with 12 bit values, we need code to get a 12 bit value 
;based on an index
;
;Index 0 = 8 bits of offset 0, 4 high bits of offset 1
;Index 1 = 4 low bits of offset 1, 8 bits of offset 2
;Index 2 = 8 bits of offset 3, 4 high bits of offset 4
;etc....

	push eax
	push esi

	mov eax, Index
	mov edx, eax
	shr edx, 1	;/2
	pushfd
	add eax, edx	;Index = Index * 1.5
	mov esi, Source
	add esi, eax
	popfd

	.if CARRY?		;value = low 4 bits, full 8 bits 
		mov ah, byte ptr[esi]
		mov al, byte ptr[esi+1]
		and eax, 111111111111b
	.else			;value = full 8 bits, 4 high bits
		mov ah, byte ptr[esi]
		mov al, byte ptr[esi+1]
		shr ax, 4
		and eax, 111111111111b
	.endif

	mov edx, eax
	pop esi
	pop eax
	ret

ReadByIndex endp

;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
	WriteByIndex proc, Destination: dword, Index: dword
;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
;Writes lower 12 bits of eax

;Writing counterpart of ReadByIndex

	push ecx
	push edx
	push esi

	mov ecx, eax
	mov eax, Index
	mov edx, eax
	shr edx, 1	;/2
	pushfd
	add eax, edx	;Index = Index * 1.5
	mov esi, Destination
	add esi, eax
	popfd

	.if CARRY?		;value = low 4 bits, full 8 bits 
		mov al, byte ptr [esi]
		and al, 11110000b	;clear low 4
		or al, ch		;set lower 4
		mov byte ptr[esi], al
		mov byte ptr[esi+1], cl	

	.else			;value = full 8 bits, 4 high bits
		mov ax, cx
		shr ax, 4	;al = upper 8 bits 
		mov byte ptr[esi], al
		mov al, byte ptr[esi+1]	
		and al, 00001111b	;clear upper 4
		mov ah, cl
		shl ah, 4		;mov lower 4 to upper 4
		or al, ah		;set upper 4
		mov byte ptr[esi+1], al
	.endif

	pop esi
	pop edx
	pop ecx
	ret	

WriteByIndex endp


